# -*- coding: utf-8 -*-
#
#  This file is part of sysscope.
#
#  sysscope - Tool for visual representation of RRDtool's Round Robin Databases.
#
#  Project: https://www.codetrax.org/projects/sysscope
#
#  Copyright 2008 George Notaras <gnot [at] g-loaded.eu>, CodeTRAX.org
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import re

from sysscope.core.exceptions import sysscopeError

#
# Constants
#

# Legend Columns
VALID_LEGEND_COLUMNS = ['Last', 'Min', 'Avg', 'Max', '95th']
DEFAULT_LEGEND_COLUMNS = VALID_LEGEND_COLUMNS

# Log Levels
VALID_LOG_LEVELS = ['debug', 'info', 'warning', 'error', 'critical']

#
# Custom Exceptions
#

class BaseSyntaxError(sysscopeError):
	def __init__(self, line_no, error_message):
		self.line_no = line_no
		self.value = error_message
		self.msg = 'syntax error in line %d. %s' % (self.line_no, self.value)
	def __str__(self):
		return self.msg

class GeneralSyntaxError(BaseSyntaxError):
	def __str__(self):
		return 'General Error: "%s"' % self.msg

class TagSyntaxError(BaseSyntaxError):
	def __str__(self):
		return 'Tag Error: "%s"' % self.msg

class DirectiveSyntaxError(BaseSyntaxError):
	def __str__(self):
		return 'Directive Error: "%s"' % self.msg

#
# Tag classes
#

class BaseTag:
	"""The base of Tag classes.

	Each tag has the following attributes:

	line		: the line number at which the directive exists in the config
	type		: 'tag'
	name		: the name of the tag
	value		: the data
	is_closing_tag	: boolean
	"""
	def __init__(self, name, data, is_closing_tag):
		self.line = 0
		self.type = 'tag'
		self.name = name
		self.value = None
		self.is_closing_tag = is_closing_tag
	def get_value(self, data):
		return data

class TagSimple(BaseTag):
	"""
	Syntax:
		<Tag>
	"""
	pass

class TagSpecial(BaseTag):
	"""
	Syntax:
		<Tag This is the data>
	"""
	def __init__(self, name, data, is_closing_tag):
		BaseTag.__init__(self, name, data, is_closing_tag)
		self.value = self.get_value(data)
	def get_value(self, data):
		if not self.is_closing_tag and not data.strip():
			raise TagSyntaxError('Invalid tag syntax')
		return data

#
# Directive
#

class BaseDirective:
	def __init__(self, name, data, syntax):
		"""
		self.data is set by derived classes by using the get_value() method.
		"""
		self.line = 0
		self.type = 'directive'
		self.name = name
		self.data = data
		self.var = self.get_var()
		self.value = self.get_value(syntax)	# --> CAN BE A STRING OR A LIST ACCORDING TO THE SYNTAX OF THE DIRECTIVE
	def get_var(self):
		"""Override"""
		pass
	def get_value(self, syntax):
		"""May be overridden"""
		if not syntax:
			return self.data	# Returns STRING
		m = re.match(syntax, self.data)
		if m:
			return list(m.groups())	# Returns LIST
		else:
			raise DirectiveSyntaxError('Invalid directive syntax')

class DirectiveSimple(BaseDirective):
	"""
	Syntax:
		Directive data
	"""
	pass

class DirectiveSpecial(BaseDirective):
	"""
	Syntax:
		Directive variable_name data
	"""
	def get_var(self):
		"""Sets the variable_name"""
		directive_data = self.data.split(' ', 1)
		if len(directive_data) >= 2:
			self.data = directive_data[1]
			return directive_data[0]
		else:
			raise DirectiveSyntaxError('Invalid directive syntax')

#
# Value validator Classes
#

class ValueValidatorBase:
	def __init__(self, valid_values):
		"""
		valid_values : list
		"""
		self.valid_values = valid_values
	def verify(self):
		"""override"""
		pass

class ValueValidatorSingle(ValueValidatorBase):
	def verify(self, value):
		"""
		value : string
		"""
		if value not in self.valid_values:
			return False
		return True

class ValueValidatorList(ValueValidatorBase):
	def verify(self, values):
		"""
		values : string (contains the comma separated list
		"""
		values_list = [val.strip() for val in values.split(',') if val.strip()]
		for value in values_list:
			if value not in self.valid_values:
				return False
		return True

#
# Main Syntax Classes
#

class Syntax(dict):
	"""Master Syntax Class.

	Basic Syntax
	------------
	The syntax of the sysscope configuration file is an Apache-style
	syntax. It consists of statements:

		Directive data

	or

		<Tag blah blah>
		...
		</Tag>


	Modular Configuration File
	--------------------------
	The configuration file may contain the directive 'Include'. This
	directive is not part of the actual sysscope configuration file syntax.
	Its syntax is:

		Include /path/to/file.conf

	Whenever it is met, it causes the config parser to include
	/path/to/file.conf

	It may be placed anywhere within the configuration file.


	Directive Placement
	-------------------
	Directives may be placed within a block defined by an opening tag (eg.
	<Tag blah>) and a closing tag (eg. </Tag>) and have an effect within the
	specific configuration block.
	Directives may also be placed outside <Tag>...</Tag> blocks. In such a
	case the directives belongs to the 'GLOBAL' scope and have effect
	throughout the configuration file.


	Tag Types
	---------
	Thre are two types of tags:

	1. Simple Tags: <Tag>...</Tag>
	2. Special Tags: <Tag blah blah>...</Tag>


	Directive Types
	---------------
	There two types of directives:

	1. Simple directives: consist of two parts:
		a) the directive name
		b) the data:
	   Example:
	   	Directive data

	2. Special directives: consist of three parts:
		a) the directive name
		b) a variable name, to which the data is assigned. The variable
		   can then be used by subsequent directives.
		c) the data
	   Example:
	   	Directive variable_name data


	Miscellaneous Info
	------------------
	Lines beginning with the '#' character (without quotes) are considered
	as comments.

	Empty lines mean nothing.

	Leading spaces mean nothing.


	Class Structure
	---------------
	This class is a dictionary (dictA) containing all the syntax details.

	The 'keys' of the dictionary should be the names of all the valid
	tag names without any of the characters <,>,/. A special tag name
	is 'GLOBAL' which represents the 'root namespace' of the configuration
	file. For example:

		self['GLOBAL'] = {...}
		self['Tag1'] = {...}
		...
		self['TagN'] = {...}

	Each key's value is another dictionary (dictB) containing those tag and
	directive names that are valid within the key's namescpace.
	dictB's syntax is:

		{
		'tag1'		: ...,
		'directive1'	: ...,
		'directive2'	: ...,
		}

	The value of each of the dictB keys varies according to the type (tag or
	directive) of the key.

	For tags, the value is just the relevant tag class name.

	For directives, the value is a tuple consisting of three items:

		( item1, item2, item3 )

		item1 : is the name of the relevant directive class.
		item2 : is a regular expression, which matches the format of the
		        directive's data
		item3 : is an *instance* of a validator class. The list of the
		        valid values for the directive is passed to the
		        constructor of the validator.

	The format of the dictionary is outlined below for tags (a) and
	directives (b).

	(a) 'Tag' : TagSpecial

	(b) 'Directive' : (
			directive_class,
			value_syntax_using_regex OR '' (if empty),
			ValidatorClass( [valid_value_1, valid_value_2, ...] ) OR None,
			)
	"""

	def __init__(self):
		"""Master Syntax Instance Constructor"""

		self['GLOBAL'] = {
			'Hostname'	: ( DirectiveSimple, '', None ),
			'Include'	: ( DirectiveSimple, '', None ),
			'OutputDir'	: ( DirectiveSimple, '', None ),
			'Logfile'	: ( DirectiveSimple, '', None ),
			'LogLevel'	: ( DirectiveSimple, '', ValueValidatorList(VALID_LOG_LEVELS) ),
			'OptionCacheDir': ( DirectiveSimple, '', None ),
			'Template'	: TagSpecial,
			'Section'	: TagSpecial,
			}

		self['Template'] = {
			'Include'	: ( DirectiveSimple, '', None ),
			'Option'	: ( DirectiveSimple, '', None ),
			'RawOption'	: ( DirectiveSimple, '', None ),
			}

		self['Section'] = {
			'Include'	: ( DirectiveSimple, '', None ),
			'RRDBase'	: ( DirectiveSimple, '', None ),
			'Graph'		: TagSpecial,
			}

		self['Graph'] = {
			'Include'	: ( DirectiveSimple, '', None ),
			'TemplateName'	: ( DirectiveSimple, '', None ),
			'Legend'	: ( DirectiveSimple, '', ValueValidatorSingle(['On', 'Off']) ),
			'LegendColumns'	: ( DirectiveSimple, '', ValueValidatorList(VALID_LEGEND_COLUMNS) ),
			'LegendFormat'	: ( DirectiveSimple, '', None ),
			'VerticalLabel'	: ( DirectiveSimple, '', None ),
			'Option'	: ( DirectiveSimple, '', None ),
			'RawOption'	: ( DirectiveSimple, '', None ),
			'DataSource'	: ( DirectiveSpecial, '(.*)::(.*)::(.*)', None ),
			'VDataSource'	: ( DirectiveSpecial, '', None ),
			'Consolidate'	: ( DirectiveSpecial, '', None ),
			'DrawGraph'	: ( DirectiveSpecial, '(.*)::(.*)::(.*)', None ),
			'Gprint'	: ( DirectiveSimple, '(.*)::(.*)', None ),
			'Comment'	: ( DirectiveSimple, '', None ),
			'Hrule'		: ( DirectiveSimple, '', None ),
			'Vrule'		: ( DirectiveSimple, '', None ),
			}

		# Open_Tags
		# When an opening tag is first met, it is added to this list.
		# When the last tag of the list is closed, then it is removed.
		# Initially the list contains the root tag 'GLOBAL'.
		self.Open_Tags = ['GLOBAL']

	def parse_line(self, line_no, line):
		"""Parses 'line' and returns an instance of the directive or tag.

		'line' should be already stripped.

		entity entity_data

		tag_name tag_data
		directive_name directive_data
		directive_name variable_name directive_data

		"""
		#print "-"*40
		#print "Processing:", line
		#print self.Open_Tags
		#print "-"*40
		# Preliminary check if line contains an opening/closing Tag
		is_tag = False
		is_tag_closing = False
		if line.startswith('<') and line.endswith('>'):
			is_tag = True
			line = line.strip('<>')
		if line.startswith('/'):
			is_tag_closing = True
			line = line.lstrip('/')

		# Split the line at the first empty space
		# This returns the entity (directive/tag) name and its associated data.
		# NOTE: data does not exist for all closing tags and for simple tags
		parts = line.split(' ', 1)
		if len(parts) == 1:
			entity, entity_data = parts[0], ''	# For 'simple tags' or 'closing tags' entity data is ''
		else:
			entity, entity_data = parts[0], parts[1]	# For 'special tags', 'simple directives', 'special directives'

		# Set the parent_tag
		# 'parent_tag' is the container for the currently processed directive or tag
		# Usually, the parent tag is the last item of the self.Open_Tags list,
		# except for the case where the line contains a closing tag.
		# In such a case, the last item of the self.Open_Tags list is deleted
		# before the parent_tag is set.
		if is_tag_closing:
			if self.Open_Tags[-1] == entity:
				del self.Open_Tags[-1]
			else:
				raise TagSyntaxError(line_no, 'Tag \'%s\' has not been previously opened' % entity)
		# Now, parent_tag is set
		parent_tag = self.Open_Tags[-1]

		# Error checking
		if not self.has_key(parent_tag) or not self[parent_tag].has_key(entity):
			error_msg = '\'%s\' not allowed in this context' % entity
			if is_tag:
				raise TagSyntaxError(line_no, error_msg)
			else:
				raise DirectiveSyntaxError(line_no, error_msg)

		# Open_Tags management
		if is_tag:
			if is_tag_closing:
				# The closing tag has already been removed from the list
				pass
			else:
				# Add the tag to the list of the currently opened tags
				self.Open_Tags.append(entity)

		#
		# Process the Tag or Directive data
		#

		# entity_data_syntax contains the syntax information, which is
		# a class name for tags or a tuple for directives
		entity_data_syntax = self[parent_tag][entity]

		# For directives, entity_data_syntax is a tuple containing syntax information
		if type(entity_data_syntax) is tuple:	# Is directive

			# if 'is_tag' *is* enabled then an error has occured
			if is_tag:
				raise TagSyntaxError(line_no, '\'%s\' is not a valid tag' % entity)

			directive_class, directive_value_syntax, value_validator = entity_data_syntax
			obj = directive_class(entity, entity_data, directive_value_syntax)
			if value_validator:
				if not value_validator.verify(obj.value):
					raise DirectiveSyntaxError(line_no, 'Invalid directive value')
			obj.line = line_no	# Set the line number in the directive obj
			return obj

		# For tags, entity_data_syntax is a class name
		else:	# Is a tag

			# if 'is_tag' *is*not* enabled then an error has occured
			if not is_tag:
				raise DirectiveSyntaxError(line_no, '\'%s\' is not a valid directive' % entity)

			tag_class = entity_data_syntax
			obj = tag_class(entity, entity_data, is_tag_closing)
			obj.line = line_no	# Set the line number in the tag obj
			return obj






# -*- coding: utf-8 -*-
#
#  This file is part of sysscope.
#
#  sysscope - Tool for visual representation of RRDtool's Round Robin Databases.
#
#  Development Web Site:
#    - http://www.codetrax.org/projects/sysscope
#  Public Source Code Repository:
#    - https://source.codetrax.org/hgroot/sysscope
#
#  Copyright 2012 George Notaras <gnot [at] g-loaded.eu>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
This file demonstrates two different styles of tests (one doctest and one
unittest). These will both pass when you run "manage.py test".

Replace these with more appropriate tests for your application.
"""

from django.test import TestCase

class SimpleTest(TestCase):
    def test_basic_addition(self):
        """
        Tests that 1 + 1 always equals 2.
        """
        self.failUnlessEqual(1 + 1, 2)

__test__ = {"doctest": """
Another way to test that 1 + 1 is equal to 2.

>>> 1 + 1 == 2
True
"""}




# -*- coding: utf-8 -*-
#
#  This file is part of sysscope.
#
#  sysscope - Tool for visual representation of RRDtool's Round Robin Databases.
#
#  Project: https://www.codetrax.org/projects/sysscope
#
#  Copyright 2008 George Notaras <gnot [at] g-loaded.eu>, CodeTRAX.org
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import rrdtool
import datetime

from sysscope.core.commons import canonical_name


class GraphOptionError(Exception):
    pass


class Graph:

    # Options without --
    NON_DASH_OPTIONS = ('DEF:', 'VDEF', 'CDEF', 'GPRINT', 'COMMENT', 'HRULE', 'VRULE')

    def __init__(self, title):

        self.Name = title
        self.CanonicalName = canonical_name(title)
        self.Template = None
        self.LegendFormat = '%9.3lf'
        self.HasLegendColumns = False

        self.RRDGraphOptions = [
            '--imgformat', 'PNG',
            '--watermark', 'Produced with sysscope',
            '--start', 'end-1d',
            '--end', 'now',
            ]

        self.extend( ['--title', title] )    # Add the title to the graph options

    def append(self, option):
        self.RRDGraphOptions.append(option)

    def extend(self, option_list):
        self.RRDGraphOptions.extend(option_list)

    def merge_options(self, user_prefs):
        """
        This is really nasty, but does what it was written for. -- Needs rewrite
        """
        # TODO
        skip_next = False
        for n, option in enumerate(user_prefs):
            if skip_next:
                skip_next = False
                continue
            elif not option.startswith('--'):
                continue
            elif option in self.RRDGraphOptions:    # If the option exists in the RRDGraphOptions (main graph options)
                option_index = self.RRDGraphOptions.index(option)    # Get the the index of the option in the RRDGraphOptions list
                pv = self.RRDGraphOptions[option_index + 1]    # possible value: This is the next item in the RRDGraphOptions list
                if not pv.startswith('--') and pv[:3] not in self.NON_DASH_OPTIONS:    # if the next item (pv) does not start with '--' and
                                                    # is not a DEF/VDEF/CDEF argument, then 'pv' is possibly the value
                                                    # of 'option'. So we need to substitute the respective item in
                                                    # the RRDGraphOptions list with the one from the user_prefs
                    if user_prefs[n+1].startswith('--') or user_prefs[n+1][:3] in self.NON_DASH_OPTIONS:
                        raise GraphOptionError('Invalid value to option %s' % option)
                    self.RRDGraphOptions[option_index + 1] = user_prefs[n+1]    # the next item in the RRDGraphOptions list
                                                    # (previously saved as pv) is replaced by the next item
                                                    # in the user_prefs list
                    #print "Replaced", option, "(", pv, ") with:", user_prefs[n+1]
                    skip_next = True    # This is set to true, so that in the next loop, option is not processed, since it has been processed
                                # in the current loop
            else:
                self.RRDGraphOptions.append(option)
                #print "Added", option
                
                # Try to get the next item of ``user_prefs``
                try:
                    nv = user_prefs[n+1]
                except IndexError:
                	# There is no next item
                    pass
                else:
                    if not nv.startswith('--') and nv[:3] not in self.NON_DASH_OPTIONS:
                        self.RRDGraphOptions.append(nv)
                        #print " with value", nv
                        skip_next = True

    def set_legend_column_titles_injection_point(self, reset=False):
        if reset:
            self.Legend_Column_Titles_Injection_Point = None
        elif self.Legend_Column_Titles_Injection_Point is not None:
            self.Legend_Column_Titles_Injection_Point = len(self.RRDGraphOptions) - 1

    def draw_graph(self, path, start='-1h', end='now'):
        """
        saves the graph at 'path'
        """
        # Merge time frame options
        self.merge_options( [ '--start', start ] )
        self.merge_options( [ '--end', end ] )
        #print repr(self.RRDGraphOptions)
        # Save the graph
        rrdtool.graph(path, *self.RRDGraphOptions)

    def get_options(self, start, end):
        """
        The same as the draw_graph method, but does not accept a path.
        It accepts starts and end times (formats acceptable by rrdtool)
        and adds them to the graph options and then returns all the options
        as a list.

        This is mainly to be used by dynamic pages.
        """
        # Merge time frame options
        self.merge_options( [ '--start', str(start) ] )
        self.merge_options( [ '--end', str(end) ] )
        # Return rrdgraph options
        return self.RRDGraphOptions

    def add_legend_column_titles(self, column_list, max_len):
        # Print graph time range
        #self.extend( [
        #    'COMMENT:From %s - %s\l' % ( 'TODO', datetime.datetime.now().strftime("%Y/%m/%d %H\:%M\:%S")),
        #    ] )
        #
        dd = 11
        legend_columns = ['COMMENT:' + ' '*max_len + ' '*2]
        comments = {
            'Last'    : 'COMMENT:' + ' '*abs(dd - len('Last')) + 'Last',
            'Min'    : 'COMMENT:' + ' '*abs(dd - len('Minimum')) + 'Minimum',
            'Avg'    : 'COMMENT:' + ' '*abs(dd - len('Average')) + 'Average',
            'Max'    : 'COMMENT:' + ' '*abs(dd - len('Maximum')) + 'Maximum',
            '95th'    : 'COMMENT:' + ' '*abs(dd - len('95th pctl')) + '95th pctl',
            }
        for column in column_list:
            #print repr(comments[column])
            legend_columns.append( comments[column] )
        # Add new line character to the last field
        legend_columns[-1] = legend_columns[-1] + '\l'
        #print legend_columns
        self.extend( legend_columns )
        self.HasLegendColumns = True

    def add_legend(self, variable, column_list):

        c2v2g = {
            'Last'    : [
                'VDEF:%slast=%s,LAST' % (variable, variable),
                'GPRINT:%slast:%s' % (variable, self.LegendFormat)
                ],
            'Min'    : [
                'VDEF:%smin=%s,MINIMUM' % (variable, variable),
                'GPRINT:%smin:%s' % (variable, self.LegendFormat)
                ],
            'Avg'    : [
                'VDEF:%savg=%s,AVERAGE' % (variable, variable),
                'GPRINT:%savg:%s' % (variable, self.LegendFormat)
                ],
            'Max'    : [
                'VDEF:%smax=%s,MAXIMUM' % (variable, variable),
                'GPRINT:%smax:%s' % (variable, self.LegendFormat),
                ],
            '95th'    : [
                'VDEF:%spct=%s,95,PERCENT' % (variable, variable),
                'GPRINT:%spct:%s\l' % (variable, self.LegendFormat)
                ]
            }
        for column in column_list:
            column_data = c2v2g[column]
            self.extend(column_data)




# -*- coding: utf-8 -*-
#
#  This file is part of sysscope.
#
#  sysscope - Tool for visual representation of RRDtool's Round Robin Databases.
#
#  Development Web Site:
#    - http://www.codetrax.org/projects/sysscope
#  Public Source Code Repository:
#    - https://source.codetrax.org/hgroot/sysscope
#
#  Copyright 2012 George Notaras <gnot [at] g-loaded.eu>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import time
import datetime

from django.contrib.auth.decorators import login_required
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.csrf import csrf_exempt
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.utils.translation import ugettext_lazy as _
from django.http import HttpResponseNotFound
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.http import HttpResponseBadRequest
from django.core.urlresolvers import reverse

from sysscope import get_version
from sysscope.models import CFG
from sysscope.utils import get_graph_times
from sysscope.core.config import Section
from sysscope.core.commons import get_graph_image_data
from sysscope.utils import resolve_time
from sysscope.utils import TIME_FRAME_POOL


@login_required
def start_view(request):
    info_dict = {
        'CFG': CFG,
        'start' : '-d',
        'timeframe': 'd',
        'version'   : get_version(),
    }
    return render_to_response(
        'sysscope/start.html', info_dict, context_instance=RequestContext(request), content_type='text/html')


@login_required
def section_view(request, section, start, timeframe):
    
    if request.method == 'POST':
        #raise Exception(repr(request.META))
        selected_timeframe = request.POST.get('selected_timeframe', None)
        backward = request.POST.get('backward', None)
        forward = request.POST.get('forward', None)
    
        if not selected_timeframe:
            selected_timeframe = timeframe
            
        # Check value
        if selected_timeframe not in TIME_FRAME_POOL.keys():
            try:
                int(selected_timeframe)
            except ValueError:
                return HttpResponseBadRequest
        
        timestamp_current = resolve_time(start)
        timestamp_from_selection = resolve_time(selected_timeframe)
            
        if backward:
            new_start = timestamp_current + timestamp_from_selection
            start = '-%d' % new_start
            timeframe = selected_timeframe
        
        elif forward:
            new_start = timestamp_current - timestamp_from_selection
            if new_start > 0:
                start = '-%d' % new_start
                timeframe = selected_timeframe
            
        return HttpResponseRedirect(reverse('section', kwargs={
            'section': section,
            'start': start,
            'timeframe': timeframe,
            }))

    # Find section object by its canonical name
    for section_obj in CFG.sections_ordered:
        if section_obj.canonical_name == section:
            break
    
    # Get graph list and break
    graph_list = CFG.graphs.get(section_obj.name, None)
    
    if not graph_list:
        return HttpResponseNotFound
    
    # Get times
    t_start, t_end, t_frame = get_graph_times(start, timeframe)
    
    info_dict = {
        'CFG': CFG,
        'section_obj': section_obj,
        'start' : start,
        'timeframe': timeframe,
        'graph_list': graph_list,
        'version'   : get_version(),
        't_start': datetime.datetime.fromtimestamp(t_start),
        't_end': datetime.datetime.fromtimestamp(t_end),
    }
    return render_to_response(
        'sysscope/section.html', info_dict, context_instance=RequestContext(request), content_type='text/html')


@login_required
def graph_archive_view(request, section, graph):
    
    # TODO: add method CFG.section_exists()
    # Find section object by its canonical name
    for section_obj in CFG.sections_ordered:
        if section_obj.canonical_name == section:
            break
    
    # Get graph list and break
    graph_list = CFG.graphs.get(section_obj.name, None)
    
    if not graph_list:
        return HttpResponseNotFound
    
    # TODO: add method CFG.graph_exists()
    graph_found = False
    for graph_obj in graph_list:
        if graph_obj.CanonicalName == graph:
            # Break! Graph found!
            graph_found = True
            break
    
    if not graph_found:
        return HttpResponseNotFound
    
    # Get times
    #time_current = time.time()
    #t_start, t_end, t_frame = get_graph_times(time_current, start, timeframe)
    
    info_dict = {
        'CFG': CFG,
        'section_obj': section_obj,
        'graph_obj': graph_obj,
        'graph_list': graph_list,
        'start' : '-d',
        'timeframe': 'd',
        'timeframe_list': ['H', 'd', 'w', 'm', 'y'],
        'version'   : get_version(),
    }
    return render_to_response(
        'sysscope/graph_archive.html', info_dict, context_instance=RequestContext(request), content_type='text/html')


@login_required
def graph_image_view(request, section, graph, start, timeframe):
    # TODO: add method CFG.section_exists()
    # Find section object by its canonical name
    for section_obj in CFG.sections_ordered:
        if section_obj.canonical_name == section:
            break
    
    # Get graph list and break
    graph_list = CFG.graphs.get(section_obj.name, None)
    
    if not graph_list:
        return HttpResponseNotFound
    
    # TODO: add method CFG.graph_exists()
    graph_found = False
    for graph_obj in graph_list:
        if graph_obj.CanonicalName == graph:
            # Break! Graph found!
            graph_found = True
            break
    
    if not graph_found:
        return HttpResponseNotFound
    
    t_start, t_end, t_frame = get_graph_times(start, timeframe)
    
    # rrdgraph options
    rrd_options = graph_obj.get_options(start=t_start, end=t_end)
    image_data = get_graph_image_data(rrd_options)
    
    response = HttpResponse(image_data, content_type='image/png')
    response['Cache-Control'] = 'max-age=0, no-store, no-cache, must-revalidate'
    return response


    

    


# -*- coding: utf-8 -*-
#
#  This file is part of sysscope.
#
#  sysscope - Tool for visual representation of RRDtool's Round Robin Databases.
#
#  Development Web Site:
#    - http://www.codetrax.org/projects/sysscope
#  Public Source Code Repository:
#    - https://source.codetrax.org/hgroot/sysscope
#
#  Copyright 2012 George Notaras <gnot [at] g-loaded.eu>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
In order to use these template tags you need to use the following in your templates

{% load sysscope_tags %}

"""

from django import template

from sysscope.utils import TIME_FRAME_POOL
from sysscope.utils import TIME_FRAMES_ORDERED
from sysscope.utils import get_section_report_title



register = template.Library()



@register.inclusion_tag('sysscope/section_links.html')
def graph_section_links(CFG, start, timeframe):
    return {
        'CFG': CFG,
        'start': start,
        'timeframe': timeframe,
        }


@register.inclusion_tag('sysscope/archive_links.html')
def graph_archive_links(CFG, section_obj, graph_obj, graph_list, start, timeframe):
    return {
        'CFG': CFG,
        'section_obj': section_obj,
        'graph_obj': graph_obj,
        'graph_list': graph_list,
        'start': start,
        'timeframe': timeframe,
        }


@register.inclusion_tag('sysscope/nav_timeframe_selector.html')
def nav_timeframe_selector(section_obj, current_timeframe):
    return {
        'section_obj': section_obj,
        'current_timeframe': current_timeframe,
        'timeframe_list': TIME_FRAMES_ORDERED,
        }


@register.inclusion_tag('sysscope/nav_timeframe_browser.html')
def nav_timeframe_browser(section_obj, current_start, current_timeframe):
    return {
        'section_obj': section_obj,
        'current_start': current_start,
        'current_timeframe': current_timeframe,
        'timeframe_list': TIME_FRAMES_ORDERED,
        }


@register.simple_tag
def section_report_title(section_obj, start, timeframe):
    return get_section_report_title(section_obj, start, timeframe)


@register.filter
def negative(value):
    return '-%s' % value


@register.filter
def timelyadverb(value):
    try:
        return TIME_FRAME_POOL[value][2]
    except KeyError:
        try:
            int(value)
        except ValueError:
            raise Exception('Invalid timeframe: %s' % repr(timeframe))
        else:
            return value


@register.filter
def friendlytime(value):
    try:
        return TIME_FRAME_POOL[value][1]
    except KeyError:
        try:
            int(value)
        except ValueError:
            raise Exception('Invalid timeframe: %s' % repr(timeframe))
        else:
            return value




# -*- coding: utf-8 -*-
#
#  This file is part of sysscope.
#
#  sysscope - Tool for visual representation of RRDtool's Round Robin Databases.
#
#  Project: https://www.codetrax.org/projects/sysscope
#
#  Copyright 2008 George Notaras <gnot [at] g-loaded.eu>, CodeTRAX.org
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import string
import subprocess

from sysscope.core.exceptions import RrdToolError


def get_variable_name_pool():
    pool = []
    for m in string.lowercase:
        for n in string.lowercase:
            pool.append("%s%s" % (m, n))
    return pool


def canonical_name(name):
    """
    - Removes disallowed characters (punctuation)
    - Replaces spaces with dashes
    - Strip dashes from start and end of string
    - Converts to lowercase
    """
    def allowed_char(c):
        if c not in string.punctuation:
            return True
    name = filter(allowed_char, name)
    name = name.replace(' ', '-')
    name = name.strip('-')
    return name.lower()


def get_graph_image_data(options):
    args = ['rrdtool', 'graph', '-']
    args.extend(options)
    #import pprint
    #pprint.pprint (args)
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    image_data = p.stdout.read()
    if not image_data:
        import pprint
        rrdtool_params = pprint.pformat(args)
        raise RrdToolError(rrdtool_params)
    errors = p.stdout.read()
    if errors:
        raise RrdToolError(errors)
    #for line in p.stderr:
    #    #print line
    #    continue    # TODO: find way to display errors: mod_wsgi by default restricts printing to stdout
    return image_data



# -*- coding: utf-8 -*-
#
#  This file is part of sysscope.
#
#  sysscope - Tool for visual representation of RRDtool's Round Robin Databases.
#
#  Development Web Site:
#    - http://www.codetrax.org/projects/sysscope
#  Public Source Code Repository:
#    - https://source.codetrax.org/hgroot/sysscope
#
#  Copyright 2012 George Notaras <gnot [at] g-loaded.eu>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import time


#
# Time related functions
#


TIME_FRAMES_ORDERED = ['S', 'M', 'H', 'd', 'w', 'm', 'y']
TIME_FRAME_POOL = {
    'S' : ('%s' % (1),            'Second', 'per second'),
    'M' : ('%s' % (60),           'Minute', 'per mnute'),
    'H' : ('%s' % (60*60),        'Hour',   'hourly'),
    'd' : ('%s' % (60*60*24),     'Day',    'daily'),
    'w' : ('%s' % (60*60*24*7),   'Week',   'weekly'),
    'm' : ('%s' % (60*60*24*30),  'Month',  'monthly'),
    'y' : ('%s' % (60*60*24*365), 'Year',   'yearly'),
}



def resolve_time(t, make_absolute=False):
    """
    Might be all supported formats
    negative notation means "relative to now"
    """
    time_current = int(time.time())
    
    t_is_relative_to_now = False
    if t.startswith('-'):
        t_is_relative_to_now = True
    
    # Strip the minus sign (if any).
    # This is not required since "t_is_relative_to_now" has been set
    t = t.lstrip('-')
    
    # If a special notation has been used, for instance: "12d",
    # resolve the time to a timestamp
    notation = t[-1]
    if notation in TIME_FRAME_POOL.keys():
        multiplier = t[:-1] or 1
        t = int(multiplier) * int(TIME_FRAME_POOL[notation][0])

    # If the time is relative to the current time, convert it to absolute time,
    # only if "make_absolute" has been set to True.
    if make_absolute:
        if t_is_relative_to_now:
            t = int(time_current) - int(t)
    
    # timestamp
    try:
        return int(t)
    except ValueError:
        raise Exception('Invalid time: %s' % repr(t))


    
def get_graph_times(start, timeframe):
    """
    Helper function.
    
    Given times in the query args might be all supported formats.
    Negative notation means "relative to now"
    
    Returns 3 integers :
        - t_start (timestamp)
        - t_end (timestamp)
        - t_frame (difference between the other two)
    
    t_start and t_end are ansolute times. This means they are timestamps
    and not times relative to the current time.
    
    """

    t_start = resolve_time(start, make_absolute=True)
    t_frame = resolve_time(timeframe, make_absolute=False)
    t_end = t_start + t_frame
    
    return t_start, t_end, t_frame


def get_section_report_title(section_obj, start, timeframe):
    """
    section: Section object
    """
    # If a special notation has been used (eg 12H), then we only need the last
    # character in order to determine which entry of the TIME_FRAME_POOL we need
    try:
        time_frame_data = TIME_FRAME_POOL[timeframe[-1]]
    except KeyError:
        try:
            return '%s timely graphs (%d seconds)' % (section_obj.name, int(timeframe))
        except ValueError:
            raise Exception('Invalid timeframe: %s' % repr(timeframe))
    else:
        return '%s %s graphs' % (section_obj.name, time_frame_data[2])




#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
#  This file is part of sysscope.
#
#  sysscope - Tool for visual representation of RRDtool's Round Robin Databases.
#
#  Development Web Site:
#    - http://www.codetrax.org/projects/sysscope
#  Public Source Code Repository:
#    - https://source.codetrax.org/hgroot/sysscope
#
#  Copyright 2012 George Notaras <gnot [at] g-loaded.eu>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  NOTES
#
#  Create source distribution tarball:
#    python setup.py sdist --formats=gztar
#
#  Create binary distribution rpm:
#    python setup.py bdist --formats=rpm
#
#  Create binary distribution rpm with being able to change an option:
#    python setup.py bdist_rpm --release 7
#
#  Test installation:
#    python setup.py install --prefix=/usr --root=/tmp
#
#  Install:
#    python setup.py install
#  Or:
#    python setup.py install --prefix=/usr
#

import sys
import os
sys.path.insert(0, os.path.abspath('src'))

from setuptools import setup

from sysscope import get_version
from sysscope import get_status_classifier

def read(fname):
    """Utility function to read the README file."""
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

if __name__=='__main__':
    setup(
        name = 'sysscope',
        version = get_version(),
        license = 'Apache License version 2',
        author = 'George Notaras',
        author_email = 'gnot [at] g-loaded.eu',
        maintainer = 'George Notaras',
        maintainer_email = 'gnot [at] g-loaded.eu',
        url = 'http://www.codetrax.org/projects/sysscope',
        description = """sysscope - Tool for visual representation of RRDtool's Round Robin Databases.""",
        long_description = read('README'),
        download_url = 'https://source.codetrax.org/hgroot/sysscope',
        platforms=['any'],
        classifiers = [
            get_status_classifier(),
            'Environment :: Web Environment',
            'Framework :: Django',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: Apache Software License',
            'Natural Language :: English',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
            'Topic :: Software Development :: Libraries :: Application Frameworks',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
        package_dir = {'': 'src'},
        packages = [
            'sysscope',
            #'sysscope.templatetags',
            #'sysscope.management',
            #'sysscope.management.commands',
        ],
        include_package_data = True,
        #install_requires=read('requirements.txt').splitlines(),
        zip_safe = False,
    )




# -*- coding: utf-8 -*-
#
#  This file is part of sysscope.
#
#  sysscope - Tool for visual representation of RRDtool's Round Robin Databases.
#
#  Project: https://www.codetrax.org/projects/sysscope
#
#  Copyright 2008 George Notaras <gnot [at] g-loaded.eu>, CodeTRAX.org
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import sys
import os
import socket

from sysscope.core.commons import canonical_name
from sysscope.core import graphbase
from sysscope.core.syntax import Syntax, DEFAULT_LEGEND_COLUMNS
from sysscope.core.exceptions import sysscopeError


class ConfigSyntaxError(sysscopeError):
	def __init__(self, line_no, error_message):
		self.line_no = line_no
		self.value = error_message
	def __str__(self):
		return 'Syntax error in line %d. %s' % (self.line_no, self.value)

class ConfigFileError(sysscopeError):
	pass


class Section:
	def __init__(self, name):
		self.name = name
		self.canonical_name = canonical_name(name)
	def __str__(self):
		return self.name


POSSIBLE_CONFIG_LOCATIONS = [
	'/usr/local/etc/sysscope.conf',
	'/etc/sysscope.conf',
	'etc/sysscope.conf'
	]

	
def get_configuration(global_only):
	"""Helper function to get the configuration.
	
	Returns a sysscope.config.ConfigFileParser object
	"""
	CONFIG_PATH = ''
	for loc in POSSIBLE_CONFIG_LOCATIONS:
		if os.path.exists(loc):
			CONFIG_PATH = loc
	CFG = ConfigFileParser(CONFIG_PATH)
	CFG.parse(global_only=global_only)
	return CFG


class ConfigFileParser:

	COMMENT_CHAR = '#'

	def __init__(self, path):

		# Configuration file path
		if not path:
			raise ConfigFileError('Configuration file not found')
		self.path = path

		# How the configuration file is parsed:
		#
		# Each line is read from the config file and is put to the
		# 'ConfigLineStack' stack. The lines are put in raw mode into
		# the stack, meaning that no processing occurs during this
		# operation, apart from striping the line and not including
		# comments or empty lines.
		#
		# A function should pop() the configuration lines from the stack
		# in order to be processed.
		#
		# This method is required in case the config file is divided
		# into several smaller config files, which are imported into
		# the main config file with the 'Include' directive.
		self.ConfigLineStack = []

		# Holds the global options, eg logfile, output dir etc
		# Format: {
		#	'option1' : value,
		#	}
		self.global_options = {}

		# Holds the tree of sections and graphs
		# Format: {
		#	'section1' : [graph1, graph2, ...],
		#	'section2' : [graph4],
		#	}
		# Types:
		#   section : str
		#   graph : graphbase.Graph instance
		#
		self.graphs = {}

		# Holds the user-defined templates
		# Format: {
		#	'template1' : ['--opt1', 'opt1_value', '--opt2', --opt3', 'opt3_value', ...],
		#	}
		self.templates = {}

		# This is set to the name of the first template that is read from the config.
		# This is by convention considered the default template.
		self.default_template = None

		# Holds Section instances in the order sections are read from the config file.
		self.sections_ordered = []

		# 'reading_graph_options' signals the gathering of all the
		# next directives, until a '</Graph>' tag is met. These
		# directives are added to the following list as directive objects.
		self.graph_directives = []	# Should be reset after graph directives have been processed
		
		# Holds the box's hostname. By default, this is set to the auto-determined hostname.
		# If the global directive 'Hostname' is found, then it is replaced by that directive's value.
		self.hostname = socket.getfqdn()

		# Every time the parser enters a configuration block, it needs
		# to store some information that is relative to the current
		# configuration block or its parent blocks. This information is
		# stored into the following variables.
		self.current_section = ''	# Should be reset at </Section>
		self.current_section_rrdbase = ''	# Should be reset at </Section>
		self.current_template = ''	# Should be reset at </Template>
		self.current_graph = None	# Should be reset after graph directives have been processed

		# Switches
		# These boolean variables show what the parser is doin, eg
		# reading the template block.

		# 'processing_template' becomes True whenever directives within
		# a <Template..>...</Template> block are processed.
		self.processing_template = False	# Should be reset at </Template>

		# 'reading_graph_options' becomes True whenever a 'Graph' tag is met.
		self.reading_graph_options = False	# Should be reset at </Graph>

		# Read main config file
		self.read_config_file(path)


	def read_config_file(self, path):
		f = open(path)
		conflines = []
		for line in f:
			line = line.strip()
			# Empty lines and comments are excluded:
			if line and not line.startswith(self.COMMENT_CHAR):
				conflines.append(line)
		f.close()
		self.ConfigLineStack.extend( reversed(conflines) )


	def parse(self, global_only=False):
		"""Parses the config file.

		if 'global_only' is True, then only the directives in the GLOBAL
		namespace are parsed. Nothing else.
		"""

		# Counter that is used for line numbering.
		no = 0

		# Master syntax class
		# the parse_line() method returns an object of type 'tag' or 'directive'
		ConfSyntax = Syntax()

		# Iterate over the stack items.
		while self.ConfigLineStack:

			# retrieve one configuration line from the stack
			line = self.ConfigLineStack.pop()

			# Increment counter by one
			#
			# Special Note:
			# The lines inside a <Graph ..>...</Graph> block are first saved in the
			# 'self.graph_directives' list along with each line number (phase A)
			# and then they are processed (phase B).
			# In phase B (graph-directive processing mode) the counter should not be increased.
			if self.graph_directives and not self.reading_graph_options:
				pass
			else:
				no += 1

			# Get entity
			obj = ConfSyntax.parse_line(no, line)

			# Check for GLOBAL ONLY PARSING
			# If global_only parsing has been requested and the read
			# object is a tag or it is an 'Include' directive,
			# do nothing.
			if global_only:
				if obj.type == 'tag':
					continue
				elif obj.name == 'Include':
					continue

			# Include directive
			# This directive imports an additional configuration file into self.ConfigLineStack
			if obj.name == 'Include':

				path = obj.value
				self.read_config_file(path)
				continue

			if obj.type == 'tag':

				if obj.is_closing_tag:	# The line contains a CLOSING TAG

					# Reset the following variables
					if obj.name == 'Section':
						self.current_section = ''
						self.current_section_rrdbase = ''
					elif obj.name == 'Template':
						self.current_template = ''
						self.processing_template = False
					elif obj.name == 'Graph':
						self.reading_graph_options = False

				else:			# The line contains an OPENING TAG

					# Process Tags

					if obj.name == 'Section':
						section_name = obj.value

						# Set the section that is being processed
						self.current_section = section_name
						# Add the section to the list, which contains the order of the sections
						self.sections_ordered.append( Section(section_name) )
						# Add the new section in the graphs dictionary
						self.graphs[section_name] = []

					elif obj.name == 'Template':
						template_name = obj.value

						# Set the current_template
						self.current_template = template_name
						# Add the template to self.templates
						self.templates[template_name] = []
						# Set the default template if it hasn't been set.
						# This way the default template is always the first one that is read from the config
						if not self.default_template:
							self.default_template = template_name
						# Set template-processing mode
						self.processing_template = True

					elif obj.name == 'Graph':
						#graph_name = obj.value
						graph_name = '%s - %s' % (self.hostname, obj.value)

						# Create instance of graph...
						GraphRRD = graphbase.Graph(graph_name)
						# ... add it to the current section's list graphs
						self.graphs[self.current_section].append( GraphRRD )
						# ... set the currently processed graph to 'GraphRRD'
						self.current_graph = GraphRRD
						# ... set graph-directive-gathering mode
						self.reading_graph_options = True

			elif self.processing_template:	# Template block processing mode

				# Process directives within the <Template ...> ... </Template> block.
				directive, directive_data = obj.name, obj.value

				if directive in ('Option', 'RawOption'):

					parts = directive_data.split(' ', 1)
					# Add the option
					if directive == 'Option':
						option = '--%s' % parts[0]
					else:
						option = parts[0]
					self.templates[self.current_template].append(option)
					# Add the value, if applicable
					if len(parts) > 1:
						option_value = parts[1]
						self.templates[self.current_template].append(option_value)

			elif self.reading_graph_options:	# Graph-directive gathering mode

				# This should *READ* lines that are within the <Graph ...> ... </Graph> block.
				# The directive object is added to the 'self.graph_directives' list
				self.graph_directives.append( obj )

			elif obj.type == 'directive':	# If it is any other directive __outside__ <Graph>...</Graph> or <Template>...</Template>

				directive, directive_data = obj.name, obj.value
				#print 'Processing', directive, directive_data
				self.global_options[directive] = directive_data
				if directive == 'Hostname':
					self.hostname = directive_data
				elif directive == 'RRDBase':
					# Replace the @Hostname@ variable
					directive_data = directive_data.replace('@Hostname@', self.hostname)
					self.current_section_rrdbase = directive_data

			if self.graph_directives and not self.reading_graph_options:		# Graph-directive processing mode

				# This should *PROCESS* self.current_graph's SAVED directives
				# (lines that have been read from a <Graph ...> ... </Graph> block)

				# Default template
				# As default template is considered the the first <Template..>...</Template> block
				# If a template is not defined for a graph using the 'TemplateName' directive, then
				# the default template is used.
				if not sum(1 for dctv in self.graph_directives if dctv.name == 'TemplateName'):
					# If TemplateName is not used in the current graph, use the default template
					#print "Using DEFAULT Template %s" % self.default_template
					self.current_graph.merge_options( self.templates[self.default_template] )

				# Process Legend-related directives.

				# DoLegend
				# By default, each 'DrawGraph' variable is added to the legend
				# and 'Last','Min','Avg','Max','95th' items are displayed.
				# The following gets False when a 'Legend off' directive is met.
				DoLegend = True
				for dctv in self.graph_directives:
					if dctv.name == 'Legend' and dctv.value == 'Off':
						DoLegend = False

				# legend_columns
				# Determine legend columns
				if DoLegend:
					legend_columns = DEFAULT_LEGEND_COLUMNS
					for dctv in self.graph_directives:
						if dctv.name == 'LegendColumns':
							# Get a list of legend columns, eg ['Last', 'Min', 'Avg', 'Max']
							legend_columns = [col.strip() for col in dctv.value.split(',') if col.strip()]

				# longest_legend_label_len
				# Determine the length of the longest legend label.
				if DoLegend:
					longest_legend_label_len = 0
					for dctv in self.graph_directives:
						if dctv.name == 'DrawGraph':
							legend_label = dctv.value[-1]
							if len(legend_label) > longest_legend_label_len:
								longest_legend_label_len = len(legend_label)

				# Process other directives
				for graph_directive in self.graph_directives:

					# Store the saved line number and the saved directive value in new variables.
					# 'saved_no' should be used in error messages in this mode. This represents
					# the actual config line number. 'no' remains unchanged in this mode.
					directive = graph_directive.name
					directive_data = graph_directive.value
					saved_no = graph_directive.line

					if directive == 'TemplateName':
						#print repr(self.templates[directive_data])
						# If the default template has been used, it is impossible to enter in here.
						self.current_graph.merge_options( self.templates[directive_data] )

					elif directive == 'Legend':
						if directive_data == 'Off':
							self.current_graph.append('--no-legend')
							# DoLegend has already been set to False

					elif directive == 'LegendColumns':
						# Already processed
						pass

					elif directive == 'LegendFormat':
						self.current_graph.LegendFormat = directive_data

					elif directive == 'VerticalLabel':
						self.current_graph.extend( ['--vertical-label', directive_data] )

					elif directive == 'Option':
						rrdgraph_option = '--%s' % directive_data
						self.current_graph.merge_options(rrdgraph_option.split(' ', 1))

					elif directive == 'RawOption':
						rrdgraph_option = directive_data
						self.current_graph.merge_options(rrdgraph_option.split(' ', 1))

					elif directive == 'DataSource':
						# Syntax:
						#   DataSource <variable_name> </path/to/file.rrd>::<DS>::<RRA>

						# Set new variables for the  contents of the directive
						variable = graph_directive.var
						rrd_path, ds, rra = directive_data

						# Translate RRDBase, if it exists in the rrd path
						rrd_path = rrd_path.replace('@RRDBase@', self.current_section_rrdbase)

						# Construct rrdtool DEF statement and add it to the current graph
						RRD_DEF = 'DEF:%s=%s:%s:%s' % (variable, rrd_path, ds, rra)
						self.current_graph.append(RRD_DEF)

					elif directive == 'Consolidate':
						# Syntax:
						#   Consolidate <variable_name> <RPN_expression>

						# Set new variables for the  contents of the directive
						variable = graph_directive.var
						variable_data = directive_data

						# Construct rrdtool CDEF statement and add it to the current graph
						RRD_CDEF = 'CDEF:%s=%s' % (variable, variable_data)
						self.current_graph.append(RRD_CDEF)

					elif directive == 'DrawGraph':
						# Syntax:
						#   DrawGraph <variable_name> <type>::<color>::<legend_label>

						# Add legend column titles.
						if DoLegend and not self.current_graph.HasLegendColumns:
							self.current_graph.add_legend_column_titles(legend_columns, longest_legend_label_len)

						# Set new variables for the  contents of the directive
						variable = graph_directive.var
						graph_type, graph_color, var_label = directive_data

						# Hidden feature:
						# If the graph is a line (LINE{1,2,3}), but the word AREA has
						# concatenated, eg LINE1AREA, then an area of the same color but with
						# an alpha of '33' is added. this produces a sexier graph.
						is_sexy = False	# get True if eg LINE1AREA has been requested.
						if graph_type.startswith('LINE') and graph_type.endswith('AREA'):
							graph_type = graph_type[:-4]	# Keep the proper line type
							is_sexy = True

						# Color can be empty in order to prevent the graph item from being drawn.
						# This is needed because if the user does not want to enter a color,
						# he MUST leave an empty space between the '::' separators.
						graph_color = graph_color.strip()

						if DoLegend:
							var_label = var_label + ' '*(longest_legend_label_len-len(var_label))
						RRD_GRAPH = '%s:%s%sFF:%s' % (graph_type, variable, graph_color, var_label)
						self.current_graph.append(RRD_GRAPH)

						# Also add the extra AREA if 'LINEXAREA' had been requested
						if is_sexy:
							RRD_GRAPH = '%s:%s%s33:%s' % ('AREA', variable, graph_color, '')
							self.current_graph.append(RRD_GRAPH)
						# Also, add the legend.
						if DoLegend:
							self.current_graph.add_legend(variable, legend_columns)

					elif directive == 'VDataSource':
						# Syntax:
						#   VDataSource <variable_name> <RPN_expression>

						# Set new variables for the  contents of the directive
						variable = graph_directive.var
						variable_data = directive_data

						# Construct rrdtool CDEF statement and add it to the current graph
						RRD_CDEF = 'VDEF:%s=%s' % (variable, variable_data)
						self.current_graph.append(RRD_CDEF)

					elif directive == 'Gprint':
						# Syntax:
						#   Gprint <variable>::<format>

						RRD_GRAPH = 'GPRINT:' + ':'.join(directive_data)	# directive_data is a list
						self.current_graph.append(RRD_GRAPH)

					elif directive == 'Comment':
						# Syntax:
						#   Comment <data>

						RRD_GRAPH = 'COMMENT:' + directive_data	# directive_data is a string for this directive
						self.current_graph.append(RRD_GRAPH)

					elif directive == 'Hrule':
						# Syntax:
						#   Hrule <data>

						RRD_GRAPH = 'HRULE:' + directive_data	# directive_data is a string for this directive
						self.current_graph.append(RRD_GRAPH)

					elif directive == 'Vrule':
						# Syntax:
						#   Vrule <data>

						RRD_GRAPH = 'VRULE:' + directive_data	# directive_data is a string for this directive
						self.current_graph.append(RRD_GRAPH)

				# All graph directives in 'self.graph_directives' have been processed.

				# 'self.current_graph' & 'self.graph_directives' are reset here
				self.current_graph = None
				self.graph_directives = []


